# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
"""All the action we need during build"""

from __future__ import annotations

import json
import os
import pathlib
import urllib.request as url_lib

import nox


def _install_bundle(session: nox.Session) -> None:
    session.run(
        "rm",
        "-rf",
        "./bundled/libs",
        external=True,
    )
    session.install(
        "-v",
        "--target",
        "./bundled/libs",
        "..",
    )


def _check_files(names: list[str]) -> None:
    root_dir = pathlib.Path(__file__).parent
    for name in names:
        file_path = root_dir / name
        lines: list[str] = file_path.read_text().splitlines()
        if any(line for line in lines if line.startswith("# TODO:")):
            raise Exception(f"Please update {os.fspath(file_path)}.")


def _update_pip_packages(session: nox.Session) -> None:
    session.run(
        "uv",
        "pip",
        "compile",
        "--generate-hashes",
        "--resolver=backtracking",
        "--upgrade",
        "./requirements.in",
    )
    session.run(
        "uv",
        "pip",
        "compile",
        "--generate-hashes",
        "--resolver=backtracking",
        "--upgrade",
        "./src/test/python_tests/requirements.in",
    )


def _get_package_data(package):
    json_uri = f"https://registry.npmjs.org/{package}"
    with url_lib.urlopen(json_uri) as response:
        return json.loads(response.read())


def _update_npm_packages(session: nox.Session) -> None:
    pinned = {
        "vscode-languageclient",
        "@types/vscode",
        "@types/node",
    }
    package_json_path = pathlib.Path(__file__).parent / "package.json"
    package_json = json.loads(package_json_path.read_text(encoding="utf-8"))

    for package in package_json["dependencies"]:
        if package not in pinned:
            data = _get_package_data(package)
            latest = "^" + data["dist-tags"]["latest"]
            package_json["dependencies"][package] = latest

    for package in package_json["devDependencies"]:
        if package not in pinned:
            data = _get_package_data(package)
            latest = "^" + data["dist-tags"]["latest"]
            package_json["devDependencies"][package] = latest

    # Ensure engine matches the package
    if (
        package_json["engines"]["vscode"]
        != package_json["devDependencies"]["@types/vscode"]
    ):
        print(
            "Please check VS Code engine version and @types/vscode version in package.json."
        )

    new_package_json = json.dumps(package_json, indent=4)
    # JSON dumps uses \n for line ending on all platforms by default
    if not new_package_json.endswith("\n"):
        new_package_json += "\n"
    package_json_path.write_text(new_package_json, encoding="utf-8")
    session.run("npm", "install", external=True)


def _setup_template_environment(session: nox.Session) -> None:
    session.install("wheel", "pip-tools")
    session.run("uv", "sync", "--active", "--group", "vscode")
    _install_bundle(session)


@nox.session()
def setup(session: nox.Session) -> None:
    """Sets up the template for development."""
    _setup_template_environment(session)


@nox.session()
def tests(session: nox.Session) -> None:
    """Runs all the tests for the extension."""
    session.run("pytest", "-svv", "src/test/python_tests")


@nox.session()
def lint(session: nox.Session) -> None:
    # check typescript code
    session.run("npm", "run", "lint", external=True)


@nox.session()
def build_package(session: nox.Session) -> None:
    """Builds VSIX package for publishing."""
    _check_files(["README.md", "LICENSE", "SECURITY.md", "SUPPORT.md"])
    _setup_template_environment(session)
    session.run("npm", "install", external=True)
    session.run("npm", "run", "vsce-package", external=True)


@nox.session()
def update_packages(session: nox.Session) -> None:
    """Update pip and npm packages."""
    session.install("wheel", "pip-tools")
    _update_pip_packages(session)
    _update_npm_packages(session)



from __future__ import annotations

import json
from dataclasses import asdict, dataclass, field
from json.decoder import JSONDecodeError
from typing import TYPE_CHECKING
from urllib import error, request

from tach.constants import GAUGE_API_BASE_URL
from tach.modularity import (
    Module,
    Usage,
    build_modules,
    build_usages,
    serialize_diagnostics,
)

if TYPE_CHECKING:
    from pathlib import Path

    import pydot  # type: ignore

    from tach.extension import ProjectConfig


@dataclass
class ShowReportMetadata:
    version: str = "1.5"


@dataclass
class ShowReport:
    modules: list[Module]
    usages: list[Usage]
    diagnostics: str
    metadata: ShowReportMetadata = field(default_factory=ShowReportMetadata)


def generate_show_report(
    project_root: Path,
    project_config: ProjectConfig,
    included_paths: list[Path],
) -> ShowReport:
    modules = build_modules(
        project_config=project_config, included_paths=included_paths
    )
    usages = build_usages(
        project_root=project_root,
        project_config=project_config,
        included_paths=included_paths,
    )
    diagnostics = serialize_diagnostics(
        project_root=project_root,
        project_config=project_config,
    )
    return ShowReport(modules=modules, usages=usages, diagnostics=diagnostics)


def upload_show_report(
    project_root: Path,
    project_config: ProjectConfig,
    included_paths: list[Path],
) -> str | None:
    show_report = generate_show_report(
        project_root=project_root,
        project_config=project_config,
        included_paths=included_paths,
    )
    json_data = json.dumps(asdict(show_report))
    json_bytes = json_data.encode("utf-8")
    req = request.Request(
        f"{GAUGE_API_BASE_URL}/api/show/graph/1.5",
        data=json_bytes,
        headers={"Content-Type": "application/json"},
    )
    try:
        # Send the request and read the response
        with request.urlopen(req) as response:
            response_data = response.read().decode("utf-8")
            response_json = json.loads(response_data)
            uid = response_json.get("uid")
            return f"{GAUGE_API_BASE_URL}/show?uid={uid}"
    except (UnicodeDecodeError, JSONDecodeError, error.URLError) as e:
        print(f"Error: {e}")
        return None


def generate_module_graph_dot_string(
    project_config: ProjectConfig,
    included_paths: list[Path],
) -> str:
    # Local import because networkx takes about ~100ms to load
    import networkx as nx

    graph = nx.DiGraph()  # type: ignore

    def upsert_edge(
        graph: nx.DiGraph,  # type: ignore
        module: str,
        dependency: str,
        dashed: bool = False,
    ) -> None:
        if module not in graph:
            graph.add_node(module)  # type: ignore
        if dependency not in graph:
            graph.add_node(dependency)  # type: ignore
        if dashed:
            graph.add_edge(module, dependency, style="dashed")  # type: ignore
        else:
            graph.add_edge(module, dependency)  # type: ignore

    modules = project_config.filtered_modules(included_paths)

    for module in modules:
        for dependency in module.depends_on or []:
            upsert_edge(graph, module.path, dependency.path, dependency.deprecated)  # type: ignore

    pydot_graph: pydot.Dot = nx.nx_pydot.to_pydot(graph)  # type: ignore
    return str(pydot_graph.to_string())  # type: ignore


def generate_module_graph_mermaid_string(
    project_config: ProjectConfig,
    included_paths: list[Path],
) -> str:
    modules = project_config.filtered_modules(included_paths)
    edges: list[str] = []
    isolated: list[str] = []
    LINE_ARROW = "-->"
    DOTTED_ARROW = "-.->"
    for module in modules:
        for dependency in module.depends_on or []:
            module_name = module.path.strip("<>")
            dependency_name = dependency.path.strip("<>")
            if dependency.deprecated:
                edges.append(f"    {module_name} {DOTTED_ARROW} {dependency_name}")
            else:
                edges.append(f"    {module_name} {LINE_ARROW} {dependency_name}")
        if not module.depends_on:
            isolated.append(f"    {module.path.strip('<>')}")

    return "graph TD\n" + "\n".join(edges) + "\n" + "\n".join(isolated)


def generate_module_graph_dot_file(
    project_config: ProjectConfig,
    output_filepath: Path,
    included_paths: list[Path],
) -> None:
    dot_data = generate_module_graph_dot_string(project_config, included_paths)
    output_filepath.write_text(dot_data)


def generate_module_graph_mermaid(
    project_config: ProjectConfig,
    output_filepath: Path,
    included_paths: list[Path],
) -> None:
    mermaid_graph = generate_module_graph_mermaid_string(project_config, included_paths)
    output_filepath.write_text(mermaid_graph)


__all__ = [
    "upload_show_report",
    "generate_module_graph_dot_file",
    "generate_module_graph_mermaid",
    "generate_module_graph_dot_string",
    "generate_module_graph_mermaid_string",
]



from __future__ import annotations

import sys
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pathlib import Path


def is_interactive() -> bool:
    return sys.stdout.isatty() and sys.stderr.isatty()


class BCOLORS:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


def colorize(text: str, color_start: str, color_end: str = BCOLORS.ENDC) -> str:
    if is_interactive():
        return f"{color_start}{text}{color_end}"
    return text


def create_clickable_link(file_path: Path, line: int | None = None) -> str:
    if line is not None:
        return f"{file_path}:{line}"
    return str(file_path)
